#!/usr/bin/env python
import sys
import os
from argparse import ArgumentParser

import numpy

from glue.ligolw import ligolw, utils, lsctables
from glue.lal import LIGOTimeGPS

argp = ArgumentParser(description="Convert BayesWave information into an appropriate set of LIGO_LW tables.")
argp.add_argument('--file-to-process', help="File to process. Required.")
argp.add_argument('--output', help="XML file to outpt. Will append '.xml.gz' to input if no argument given.")
argp.add_argument('--channel', help="Name of channel, required. E.g. 'H1:FAKE-STRAIN'")
argp.add_argument('--search', default='BayesWave', help="Name of search, default is 'BayesWave'.")
args = argp.parse_args()

try:
    ifo, channel = args.channel.split(":")
except ValueError:
    sys.exit("Invalid channel specification")

proc_table = lsctables.New(lsctables.ProcessTable, ['comment', 'process_id', 'ifos', 'username', 'program'])
proc_row = proc_table.RowType()
proc_row.process_id = pid = proc_table.get_next_id()
proc_row.ifos = ifo
proc_row.program = sys.argv[0]
proc_row.username = os.getlogin()
proc_row.comment = "converted from BayesWave output"
proc_table.append(proc_row)

tbl = lsctables.New(lsctables.SnglBurstTable, ["ifo", "peak_time", 
      "peak_time_ns", "start_time", "start_time_ns",
      "duration",  "search", "event_id", "process_id",
      "central_freq", "channel", "amplitude", "snr", "confidence",
      "chisq", "chisq_dof", "bandwidth"])

#t_c, f_c, dur, band, snr = numpy.loadtxt(args.file_to_process, unpack=True)
rows = numpy.loadtxt(args.file_to_process)

#for t_c, f_c, dur, band, snr in rows:
for map_D, bf, snr, t_c, dur, f_c, band in rows:
  row = tbl.RowType()
  # Assumption: box centered on (t_c, f_c) with width (dur, band)
  row.set_peak(LIGOTimeGPS(float(t_c)))
  row.set_start(row.get_peak() - float(dur)/2.0)
  row.duration = dur
  row.bandwidth = band
  row.central_freq = f_c

  row.chisq = 0
  row.chisq_dof = 2*band*dur
  row.confidence = bf/(bf+1) # FIXME: Do we care to do this right?

  row.snr = snr
  row.amplitude = 1.0 # FIXME: hrss?

  row.ifo = ifo
  row.channel = channel
  row.search = args.search
  row.event_id = tbl.get_next_id()
  row.process_id = pid

  tbl.append(row)

# Build output document
xmldoc = ligolw.Document()
xmldoc.appendChild(ligolw.LIGO_LW())
xmldoc.childNodes[0].appendChild(proc_table)
xmldoc.childNodes[0].appendChild(tbl)

outf = args.output or args.file_to_process + ".xml.gz"
utils.write_filename(xmldoc, outf, gz=outf.endswith(".xml.gz"))
